﻿#include "common.h"

/*******************************************************************************************
 * 功能     ：	打开串口
 * port     :	串口号, 如("COM1")
 * baud_rate:	波特率
 * date_bits:	数据位（有效范围4~8）
 * stop_bit :	停止位
 * parity   :	奇偶校验。默认为无校验。NOPARITY 0； ODDPARITY 1；EVENPARITY 2
 ********************************************************************************************/
HANDLE Init_COM(LPCTSTR Port, int baud_rate, BYTE date_bits, BYTE stop_bit, BYTE parity)
{
	HANDLE Handle_Com = CreateFile(Port, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL);//同步方式打开串口
	if (INVALID_HANDLE_VALUE == Handle_Com)
	{
		return INVALID_HANDLE_VALUE;
	}
	SetupComm(Handle_Com, 4096, 4096);//设置缓存
	DCB dcb;
	if (!GetCommState(Handle_Com, &dcb))
	{
		cout << "获取串口配置失败" << endl;
	}
	dcb.BaudRate = baud_rate;	//波特率
	dcb.fBinary = TRUE;			//二进制模式。必须为TRUE
	dcb.ByteSize = date_bits;	//数据位。范围4-8
	if (stop_bit == 0)
		dcb.StopBits = ONESTOPBIT;	//停止位
	if (stop_bit == 1)
		dcb.StopBits = ONE5STOPBITS;	//停止位
	if (stop_bit == 2)
		dcb.StopBits = TWOSTOPBITS;	//停止位

	if (parity == NOPARITY)
	{
		dcb.fParity = FALSE;	//奇偶校验关闭
		dcb.Parity = parity;	//校验模式
	}
	else
	{
		dcb.fParity = TRUE;		//奇偶校验开启
		dcb.Parity = parity;	//校验模式
	}
	cout << SetCommState(Handle_Com, &dcb) << endl;
	PurgeComm(Handle_Com, PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR | PURGE_TXABORT);//清除缓存
	//设置串口读写时间
	COMMTIMEOUTS CommTimeOuts;
	GetCommTimeouts(Handle_Com, &CommTimeOuts);
	CommTimeOuts.ReadIntervalTimeout = 5;
	CommTimeOuts.ReadTotalTimeoutMultiplier = 0;
	CommTimeOuts.ReadTotalTimeoutConstant = 0;
	CommTimeOuts.WriteTotalTimeoutMultiplier = 10;
	CommTimeOuts.WriteTotalTimeoutConstant = 1000;
	if (!SetCommTimeouts(Handle_Com, &CommTimeOuts)) {
		return INVALID_HANDLE_VALUE;
	}
	return Handle_Com;
}

/*********************************************************************************************
* 功能     ：   发送响应数据
* 描述	   ：	向串口写入数据
* 返回值   ：   true  发送成功   false  发送失败
* m_hComm  ：   串口句柄
* data     ：   要写入的数据
* len      ：   写入数据的长度
**********************************************************************************************/
bool SendData(HANDLE m_hComm, char* data, int len)
{
	if (m_hComm == INVALID_HANDLE_VALUE)
		return FALSE;

	//清空串口
	PurgeComm(m_hComm, PURGE_RXCLEAR | PURGE_TXCLEAR);
	//写串口
	DWORD dwWrite = 0;
	DWORD dwRet = WriteFile(m_hComm, data, len, &dwWrite, NULL);
	//清空串口
	PurgeComm(m_hComm, PURGE_RXCLEAR | PURGE_TXCLEAR);
	if (!dwRet)
		return FALSE;
	return TRUE;
}

/*********************************************************************************************
* 功能     ：   Ms延时器
* 描述	   ：	ms级准确延时
* 输入     ：   lTime 延时的ms数
* 返回值   ：   无
**********************************************************************************************/
void MSleep(long lTime)
{
	LARGE_INTEGER litmp;
	LONGLONG QPart1, QPart2;
	double dfMinus, dfFreq, dfTim, dfSpec;
	QueryPerformanceFrequency(&litmp);
	dfFreq = (double)litmp.QuadPart;
	QueryPerformanceCounter(&litmp);
	QPart1 = litmp.QuadPart;
	dfSpec = 0.000001*lTime;

	do
	{
		QueryPerformanceCounter(&litmp);
		QPart2 = litmp.QuadPart;
		dfMinus = (double)(QPart2 - QPart1);
		dfTim = dfMinus / dfFreq;
	} while (dfTim<dfSpec);
}

/*********************************************************************************************
* 功能     ：   检查端口名称
* 描述	   ：	判断端口名称是否正确
* 输入     ：   Comm 端口名
* 返回值   ：   true 端口名正确
*				false 端口名错误
**********************************************************************************************/
bool Check_Input_COMM(string Comm)
{
	string compare1 = "com";
	string compare2 = "COM";
	for (unsigned int i = 0; i < 3; i++)
	{
		if (Comm[i] != compare1[i] && Comm[i] != compare2[i])
			return false;
	}
	if (Comm[3] == 0x30)
		return false;
	for (unsigned int i = 4; i < Comm.length(); i++)
	{
		if (Comm[i] >= 0x30 && Comm[i] <= 0x39)
			continue;
	}
	return true;
}

/*********************************************************************************************
* 功能     ：   输入端口名称
* 描述	   ：	获取串口通信的端口名称
* 输入     ：   无
* 返回值   ：   COMM 端口名
**********************************************************************************************/
string Input_COMM(void)
{
	int flage = 1;
	string Comm;
	do
	{
		if (flage == 1)
		{
			printf("请输入端口名称：");
			flage = 0;
		}
		else
			printf("请重新输入端口名称：");
		cin >> Comm;
		cin.clear();
		cin.sync();
	} while (!Check_Input_COMM(Comm));
	return Comm;
}

/*********************************************************************************************
* 功能     ：   检查波特率
* 描述	   ：	检查串口通信的波特率
* 输入     ：   Baud_Rate 波特率
* 返回值   ：   true  波特率正确
*				false 波特率错误
**********************************************************************************************/
bool Check_Input_Baud_Rate(unsigned int Baud_Rate)
{
	unsigned int Compare_Baud_Rate[9] = {300,600,1200,2400,4800,9600,14400,19200,38400};
	for (int i = 0; i < 9; i++)
	{
		if (Baud_Rate == Compare_Baud_Rate[i])
			return true;
	}
	return false;
}

/*********************************************************************************************
* 功能     ：   输入波特率
* 描述	   ：	获取串口通信的波特率
* 输入     ：   无
* 返回值   ：   Baud_Rate 波特率
**********************************************************************************************/
unsigned int Input_Baud_Rate(void)
{
	unsigned int Baud_Rate;
	int flage = 1;
	printf("支持的波特率有：300,600,1200,2400,4800,9600,14400,19200,38400 \n");
	do
	{
		if (flage == 1)
		{
			printf("请输入串口波特率：");
			flage = 0;
		}
		else
			printf("请重新输入串口波特率：");
		cin >> Baud_Rate;
		cin.clear();
		cin.sync();
	} while (!Check_Input_Baud_Rate(Baud_Rate));
	return Baud_Rate;
}


/*********************************************************************************************
* 功能     ：   输入数据位
* 描述	   ：	获取串口通信的数据位个数
* 输入     ：   无
* 返回值   ：   Date_Bits 数据位
**********************************************************************************************/
BYTE Input_Date_Bits(void)
{
	BYTE Date_Bits;
	int flage = 1;
	unsigned int data_bits = 0;
	printf("支持的数据位有：5,6,7,8 \n");
	do
	{
		if (flage == 1)
		{
			printf("请输入有效数据位：");
			flage = 0;
		}
		else
			printf("请重新输入有效数据位：");
		cin >> data_bits;
		cin.clear();
		cin.sync();
	} while (!(data_bits == 5 || data_bits == 6 || data_bits == 7 || data_bits == 8));
	Date_Bits = (BYTE)data_bits;
	return Date_Bits;
}

/*********************************************************************************************
* 功能     ：   检查停止位
* 描述	   ：	检测停止位是否和数据位相符
* 输入     ：   Date_Bits 数据位
*				Stop_Bits 停止位
* 返回值   ：   true  停止位正确
*				false 停止位错误
**********************************************************************************************/
bool Check_Input_Stop_Bits(BYTE Date_Bits, BYTE Stop_Bits)
{
	if (Stop_Bits == 0 || Stop_Bits == 1 || Stop_Bits == 2)
	{
		if (Stop_Bits == 1 && (Date_Bits == 6 || Date_Bits == 7 || Date_Bits == 8))
			return false;
		if (Stop_Bits == 2 && Date_Bits == 5)
			return false;
	}
	else
		return false;
	return true;
}

/*********************************************************************************************
* 功能     ：   输入停止位
* 描述	   ：	获取串口通信的停止位
* 输入     ：   Date_Bits 数据位    
* Stop_Bits= 0,1,2对应的是1bit,1.5bits,2bits.
* Date_Bits=6,7,8时   Stop_Bits不能为1
* Date_Bits=5时   Stop_Bits不能为2
* 返回值   ：   Stop_Bits 数据位
**********************************************************************************************/
BYTE Input_Stop_Bits(BYTE Date_Bits)
{
	BYTE Stop_Bits;
	int flage = 1;
	unsigned int stop_bits = 0;
	printf("支持的停止位有：0,1,2 \n");
	do
	{
		if (flage == 1)
		{
			printf("请输入停止位：");
			flage = 0;
		}
		else
			printf("请重新输入停止位：");
		cin >> stop_bits;
		cin.clear();
		cin.sync();
	} while (!Check_Input_Stop_Bits(Date_Bits, (BYTE)stop_bits));
	Stop_Bits = (BYTE)stop_bits;
	return Stop_Bits;
}

/*********************************************************************************************
* 功能     ：   选择校验位
* 描述	   ：	获取串口通信的校验位
* 输入     ：   无
* 返回值   ：   Parity 校验位
**********************************************************************************************/
BYTE Input_Parity(void)
{
	BYTE Parity;
	int flage = 1;
	unsigned int parity = 0;
	printf("支持的校验位有：0- 无校验 1- 奇校验 2- 偶校验 \n");
	do
	{
		if (flage == 1)
		{
			printf("请输入校验位：");
			flage = 0;
		}
		else
			printf("请重新输入校验位：");
		cin >> parity;
		cin.clear();
		cin.sync();
	} while (!(parity == 1 || parity == 0 || parity == 2));
	Parity = (BYTE)parity;
	return Parity;
}

/*********************************************************************************************
* 功能     ：   选择校验位
* 描述	   ：	获取串口通信的校验位
* 输入     ：   无
* 返回值   ：   Parity 校验位
**********************************************************************************************/
unsigned int Input_RTU_Enable(void)
{
	unsigned int RTU_Enable;
	int flage = 1;
	printf("请选择设备是否故障   0： 故障  1：未故障\n");
	do
	{
		if (flage == 1)
		{
			printf("请输入故障码：");
			flage = 0;
		}
		else
			printf("请重新输入故障码：");
		cin >> RTU_Enable;
		cin.clear();
		cin.sync();
	} while (!(RTU_Enable == 1 || RTU_Enable == 0));
	return RTU_Enable;
}

/*********************************************************************************************
* 功能    ：	计算写入数据的字节数
* 描述	  ：	通过对应的功能码和操作数量计算对应的数据字节数
* 输入    ：    Function_code 选择的功能码类型 Operations_Number对应功能码类型的操作数量
* 输出    ：    Read_date_number 写入数据的字节数
*********************************************************************************************/
unsigned int Count_Read_date_number(int Function_code, unsigned int Operations_Number)
{
	unsigned int Read_date_number = 0;
	if (Function_code == 0x01 || Function_code == 0x0F) //读线圈
	{
		Read_date_number = Operations_Number / 8;
		if (Operations_Number % 8)
			Read_date_number++;
	}
	if (Function_code == 0x03 || Function_code == 0x10)//读寄存器
		Read_date_number = Operations_Number * 2;
	return Read_date_number;
}


/*********************************************************************************************
* 功能     ：   计算CRC校验
* 描述	   ：	获取Modbus—CRC-16的校验数据
* 输入     ：   *Data 计算校验数据   CRC_Len 数据长度
* 返回值   ：   Ret_CRC_date CRC校验结果
**********************************************************************************************/
UINT16 CRC_16(UINT8 *Data, unsigned int CRC_Len)
{
	UINT16 CRC_date = 0XFFFF;//16位crc寄存器预置
	UINT16 temp;
	unsigned int i = 0, j = 0;
	for (i = 0; i < CRC_Len; i++)
	{
		temp = *Data & 0X00FF;//将八位数据与CRC寄存器亦或
		Data++;
		CRC_date ^= temp;
		for (j = 0; j < 8; j++)
		{
			if (CRC_date & 0X0001)//判断右移出的是不是1，如果是1则与多项式进行异或。
			{
				CRC_date >>= 1;
				CRC_date ^= 0XA001;
			}
			else
			{
				CRC_date >>= 1;
			}
		}
	}
	UINT16 Ret_CRC_date = CRC_date >> 8;
	Ret_CRC_date = Ret_CRC_date | CRC_date << 8;
	return Ret_CRC_date;
}

/*********************************************************************************************
* 功能    ：	日志记录
* 描述	  ：	记录每次通信的请求和响应报文
* 输入    ：    *Message 消息帧存放的数组   flage = 1 请求报文  flage = 其他 响应报文
*				Message_len 报文数据长度
* 输出    ：    无
*********************************************************************************************/

void Log_Note(UINT8 *Message, int flage, int Message_len)
{
	FILE *fp = NULL;
	time_t timep;
	struct tm *p;
	time(&timep);
	p = gmtime(&timep);
	string Recv_str = to_string(1900 + p->tm_year) + "-" + to_string(1 + p->tm_mon) + "-" + to_string(p->tm_mday)
		+ " " + to_string(8 + p->tm_hour) + ":" + to_string(p->tm_min) + ":" + to_string(p->tm_sec) + " Recv:";
	string Send_str = to_string(1900 + p->tm_year) + "-" + to_string(1 + p->tm_mon) + "-" + to_string(p->tm_mday)
		+ " " + to_string(8 + p->tm_hour) + ":" + to_string(p->tm_min) + ":" + to_string(p->tm_sec) + " Send:";

	/* 打开文件用于读写 */
	if ((fp = fopen("Modbus_TCP_log.txt", "a+")) == NULL)
	{
		printf("打开文件失败");
		return;
	}


	if (flage == 1)
	{
		fwrite(Send_str.c_str(), Send_str.length(), 1, fp);
		for (int i = 0; i < Message_len; i++)
		{
			fprintf(fp, "%02X ", Message[i]);
		}
		fprintf(fp, "\n");
	}
	else
	{
		fwrite(Recv_str.c_str(), Recv_str.length(), 1, fp);
		for (int i = 0; i < Message_len; i++)
		{
			fprintf(fp, "%02X ", Message[i]);
		}
		fprintf(fp, "\n");
	}
	fclose(fp);
	fp = NULL;
}


